# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

import bisect
from collections.abc import Iterable
from datetime import timedelta
from enum import Enum
from typing import List, Optional, Type

import astropy.units as u
import numpy as np
import numpy.typing as npt
from astropy.time import Time, TimeDelta


def is_contiguous(iterable: Iterable) -> bool:
    """
    Determine if a sortable iterable contains contiguous elements.
    Args:
        iterable: an iterable collection that can be sorted

    Returns:
        True if the elements are contiguous, and False otherwise
    """
    s = sorted(iterable)
    diffs = np.diff(s)
    return np.all(diffs == 1)


def flatten(lst):  # type: ignore
    """Flattens any iterable, no matter how irregular.
       Deliberately left untyped to allow for maximum type usage.

    Example:
        flatten([1, 2, [3, 4, 5], [[6, 7], 8, [9, 10]]])

    Args:
        lst: n-dimensional array

    Yields:
        Value of the iterable.
    """
    for el in lst:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


def round_minute(time: Time, up: bool = False) -> Time:
    """Round a time down (truncate) or up to the nearest minute time: an astropy.Time

    Args:
        time: times value(s) to round down/up
        up: bool indicating whether to round up

    Returns:
        Round up/down value(s) on Astropy Time object
    """
    t = time.copy()
    t.format = 'iso'
    t.out_subfmt = 'date_hm'
    if up:
        sec = t.strftime('%S').astype(int)
        idx = np.where(sec > 0)
        t[idx] += 1.0 * u.min
    return Time(t.iso, format='iso', scale='utc')


def str_to_bool(s: Optional[str]) -> bool:
    """Conversion from string to bolean

    Arg:
        s: parameter to convert

    Returns:
        true if and only if s is defined and some variant capitalization of 'yes' or 'true'.
    """
    return s is not None and s.strip().upper() in ['YES', 'TRUE']


# A dict of signs for conversion.
SIGNS = {'': 1, '+': 1, '-': -1}


def dmsstr2deg(s: str) -> float:
    """Degrees, minutes, seconds (in string form) to decimal degrees

    Args:
        s: string to convert

    Raises:
        ValueError: wrong format

    Returns:
        float: value in decimal degrees

    """
    if not s:
        raise ValueError(f'Illegal DMS string: {s}')

    sign = '+'
    if s[0] in SIGNS:
        sign = s[0]
        s = s[1:]

    result = s.split(':')
    if len(result) != 3:
        raise ValueError(f'Illegal DMS string: {s}')
    return dms2deg(int(result[0]), int(result[1]), float(result[2]), sign)


def dms2deg(d: int, m: int, s: float, sign: str) -> float:
    """Degrees, minutes, seconds to decimal degrees

    Args:
        d (int): Degree
        m (int): Minute
        s (float): Seconds
        sign (str): Mathematical sign

    Raises:
        ValueError: If signs is not in the SIGN dictionary

    Returns:
        float: Decimal degrees value
    """
    if sign not in SIGNS:
        raise ValueError(f'Illegal sign "{sign}" in DMS: {sign}{d}:{m}:{s}')
    dec = SIGNS[sign] * (d + m / 60.0 + s / 3600.0)
    return dec if dec < 180 else -(360 - dec)


def dms2rad(d: int, m: int, s: float, sign: str) -> float:
    """Degrees, minutes, seconds to radians

    Args:
        d (int): Degree
        m (int): Minute
        s (float): Seconds
        sign (str): Mathematical sign

    Raises:
        ValueError: If signs is not in the SIGN dictionary

    Returns:
        float: Radian value
    """
    return dms2deg(d, m, s, sign) * np.pi / 180.0


def hmsstr2deg(s: str) -> float:
    """HH:mm:ss in string to degrees

    Args:
        s (str): String to be transform

    Raises:
        ValueError: Wrong format

    Returns:
        float: Value in degrees
    """
    if not s:
        raise ValueError(f'Illegal HMS string: {s}')

    result = s.split(':')
    if len(result) != 3:
        raise ValueError(f'Illegal HMS string: {s}')

    return hms2deg(int(result[0]), int(result[1]), float(result[2]))


def hms2deg(h: int, m: int, s: float) -> float:
    """HH:mm:ss to degrees

    Args:
        h (int): Hour
        m (int): Minute
        s (float): Second

    Returns:
        float: Value in degrees
    """
    return h + m / 60.0 + s / 3600.0


def hms2rad(h: int, m: int, s: float) -> float:
    """HH:mm:ss to radians

    Args:
        h (int): Hour
        m (int): Minute
        s (float): Second

    Returns:
        float: Value in degrees
    """
    return hms2deg(h, m, s) * np.pi / 12.


def angular_distance(ra1: float, dec1: float, ra2: float, dec2: float) -> float:
    """Calculate the angular distance between two points on the sky.
    based on
    https://github.com/gemini-hlsw/lucuma-core/blob/master/modules/core/shared/src/main/scala/lucuma/core/math/Coordinates.scala#L52

    Args:
        ra1 (float): Right Ascension for point 1
        dec1 (float): Declination for point 1
        ra2 (float): Right Ascension for point 2
        dec2 (float): Declination for point 2

    Returns:
        float: Angular Distance
    """
    phi_1 = dec1
    phi_2 = dec2
    delta_phi = dec2 - dec1
    delta_lambda = ra2 - ra1
    a = np.sin(delta_phi / 2) ** 2 + np.cos(phi_1) * np.cos(phi_2) * np.sin(delta_lambda / 2) ** 2
    return 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))


def lerp(first_value: float, last_value: float, n: int) -> npt.NDArray[float]:
    """
    Perform linear interpolation for n points between start_value and end_value.

    Args:
        first_value: a float value from which to start
        last_value: a float value at which to end
        n: the number of points to interpolate over

    Returns:
        A numpy array of length n that linearly interpolates exclusively between start_value and end_value.
    """
    # Create the linspace array. In order to properly interpolate, we have to add 2 to n and then cut off the first
    # and last values, which will be first_value and last_value.
    return np.linspace(first_value, last_value, n + 2)[1:-1]


def lerp_enum(enum_class: Type[Enum], first_value: float, last_value: float, n: int) -> npt.NDArray[float]:
    """
    Given an Enum of float, a first_value, a last_value, and a number of slots, interpolate over the Enum
    to create a numpy array of slots.

    Args:
        enum_class: an Enum with float values
        first_value: a value in the Enum
        last_value: a value in the Enum
        n: the number of slots to interpolate over

    Returns:
        A numpy array of length n that linearly interpolates exclusively between first_value and last_value only
        containing values from enum_class.
    """
    # Create the linspace array. In order to properly interpolate, we have to add 2 to n and then cut off the first
    # and last values, which will be first_value and last_value.
    interp_values = np.linspace(first_value, last_value, n + 2)[1:-1]

    # Sort the Enum values.
    sorted_values = sorted(o.value for o in enum_class)

    # Interpolate over the Enum.
    if first_value <= last_value:
        result = [sorted_values[bisect.bisect_left(sorted_values, x)] for x in interp_values]
    else:
        result = [sorted_values[bisect.bisect_right(sorted_values, x)] for x in interp_values]
    return np.array(result)


def _lerp_circular(first_value: float, last_value: float, n: int, max_val: float) -> npt.NDArray[float]:
    """
    Perform linear interpolation for n points between first_value and last_value.
    As this is done for a circle, max_value must be specified for the maximum value (exclusive) of measurement.
    For degrees, max_val should be 360.
    For radians, max_val should be 2π.

    Args:
        first_value: a float value from which to start
        last_value: a float value at which to end
        n: the number of points to interpolate over
        max_val: the maximum value (exclusive) of the units of measurement

    Returns:
        A numpy array of length n that linearly interpolates exclusively between:
        (first_value % max_val) and (last_value % max_val)
        with values in the range [0, max_val), taking the shortest route around the circle.

        If the points are antipodal, and:
        * (first_value % max_value) < (last_value % max_value), the clockwise direction is taken.
        * (first value % max_value) > (last_value % max_value), the counter-clockwise direction is taken.
    """
    if max_val <= 0:
        raise ValueError(f'Circular linear interpolation requires positive maximum value, received: {max_val}.')

    # Normalize start and end values to the range [0, max_val).
    first_value %= max_val
    last_value %= max_val

    # Calculate distances.
    direct_clockwise_distance = (last_value - first_value) % max_val
    direct_counterclockwise_distance = (first_value - last_value) % max_val

    # Choose the direction with the shortest distance.
    if direct_clockwise_distance < direct_counterclockwise_distance:
        shortest_distance = direct_clockwise_distance
    else:
        shortest_distance = -direct_counterclockwise_distance

    # Generate the lerp.
    return (first_value + shortest_distance * lerp(0, 1, n)) % max_val


def lerp_radians(start_value: float, end_value: float, n: int) -> npt.NDArray[float]:
    """
    Perform linear interpolation for n points around a circle in radians.

    Args:
        start_value: a float value in radians from which to start
        end_value: a float value in radians at which to end
        n: the number of points to interpolate over

    Returns:
        A numpy array of length n that linearly interpolates exclusively between:
        (first_value % 2π) and (last_value % 2π)
        in radians, taking the shortest route around the circle.

        If the points are antipodal, i.e. at distance π, and:
        * (first_value % 2π) < (last_value % 2π), the clockwise direction is taken.
        * (first value % 2π) > (last_value % 2π), the counter-clockwise direction is taken.

    Raises:
        ValueError if max_val <= 0.
    """
    return _lerp_circular(start_value, end_value, n, 2 * np.pi)


def lerp_degrees(start_value: float, end_value: float, n: int) -> npt.NDArray[float]:
    """
    Perform linear interpolation for n points around a circle in degrees.

    Args:
        start_value: a float value in degrees from which to start
        end_value: a float value in degrees at which to end
        n: the number of points to interpolate over

    Returns:
        A numpy array of length n that linearly interpolates exclusively between:
        (first_value % 360) and (last_value % 360)
        in degrees, taking the shortest route around the circle.

        If the points are antipodal, i.e. at distance 180, and:
        * (first_value % 360) < (last_value % 360), the clockwise direction is taken.
        * (first value % 360) > (last_value % 360), the counter-clockwise direction is taken.
    """
    return _lerp_circular(start_value, end_value, n, 360.0)


def timedelta_astropy_to_python(time_delta: TimeDelta) -> timedelta:
    """
    Convert an AstroPy TimeDelta to a Python timedelta in seconds.

    Args:
        time_delta: an AstroPy TimeDelta

    Returns:
        the equivalent Python timedelta in seconds

    Raises:
        ValueError if TimeDelta is not a scalar.
    """
    if time_delta.ndim != 0:
        raise ValueError(f'time_slot_length contains multiple values: {len(time_delta.value)}')

    return timedelta(seconds=time_delta.to_value(u.s))


def time_delta_astropy_to_minutes(time_delta: TimeDelta) -> int:
    """
    Convert an AstroPy TimeDelta
    Args:
        time_delta: an AstroPy TimeDelta

    Returns:
        An int representing the number of minutes in the AstroPy TimeDelta.

    Raises:
        ValueError if TimeDelta is not a scalar in minutes.
    """
    if time_delta.ndim != 0:
        raise ValueError(f'time_slot_length contains multiple values: {len(time_delta.value)}')
    minutes = time_delta.to_value(u.minute)

    if int(minutes) != minutes:
        raise ValueError(f'time_slot_length is not a multiple of minutes: {minutes} minutes.')

    return int(minutes)


def first_nonzero_time(inlist: List[timedelta]) -> Optional[int]:
    """Find the index of the first nonzero timedelta in inlist
       Designed to work with the output from cumulative_seq_exec_times
       Args:
           inlist (List[timedelta]): a list of cumulative timedelta objects

       Returns:
          Optional[int]: return the index for the first non-zero, if it does
          not exist return None
    """
    for i, td in enumerate(inlist):
        if td != timedelta(0):
            return i
    return None


def standards_for_nir(exec_sci: timedelta,
                      wavelengths: Optional[List[float]] = None,
                      mode: str = 'spectroscopy') -> int:
    """
    Calculated the number of NIR standards from the length of the NIR science and the mode
    Args:
        exec_sci(timedelta): execution time for science
        wavelengths (List[float]): list of wavelengths values
        mode (str): mode for the NIR observation
    """

    # TODO: need mode or other info to distinguish imaging from spectroscopy
    if mode == 'imaging':
        time_per_standard = timedelta(hours=2.0)
    else:
        if wavelengths:
            if all(wave <= 2.5 for wave in wavelengths):
                time_per_standard = timedelta(hours=1.5)
            else:
                time_per_standard = timedelta(hours=1.0)
        else:
            raise ValueError("Wrong mode: spectroscopy expects a non-empty wavelengths list.")

    return max(1, int(exec_sci // time_per_standard))  # TODO: confirm this



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from typing import Optional, Tuple

import astropy.units as u
import numpy as np
from astropy.coordinates import (Angle, Distance, EarthLocation,
                                 GeocentricTrueEcliptic, SkyCoord)
from astropy.time import Time, TimeDelta

from .altitude import Altitude
from .constants import EQUAT_RAD, J2000, JYEAR, JYEAR_100
from .utils import (current_geocent_frame, geocentric_coors,
                    hour_angle_to_angle, local_sidereal_time)


class Moon:
    """Interface for night events for the moon on other calculations.

    To use this is required to chain the `at` method at the beginning.
    If not unhandled errors would happen.

    Attributes:
        time (Time): regular Time in UTC.
        time_jd (Time): Time on Julian day.
        time_ttjd (Time): Time on terrestrial time and Julian day.
        scalar_input (bool): Check if input is an array or a scalar.
        self.pie (float): Not a lemon one! Variable for location calculation.
        self.lambd (float): Variable for location calculation.
        self.beta (float): Variable for location calculation.

    """
    def __init__(self):
        # TODO: skip constructor altogether and just use the at method?
        self.time = None
        self.time_jd = None
        self.time_ttjd = None  # important to use correct time argument for this!!
        self.scalar_input = False
        self.pie = None
        self.lambd = None
        self.beta = None

    def at(self, time: Time) -> "Moon":
        """Set time values for other calculations.

           This method is meant to be called with:
            - location methods (either low or accurate).
            - time_by_altitude
            - rise_and_set

        Args:
            time (Time): Specific time.

        Returns:
            Self (Moon): Same object for method chaining.
        """
        self.time = time
        self.time_jd = np.asarray(time.jd)
        self.time_ttjd = np.asarray(time.tt.jd)  # important to use correct time argument for this!!
        self.scalar_input = False
        if self.time_ttjd.ndim == 0:
            self.time_ttjd = self.time_ttjd[None]  # Makes 1D
            self.scalar_input = True

        return self

    def _low_precision_calculations(self) -> None:
        """Compute low precision values for the moon location method calculations.
        """
        t = (self.time_jd - J2000) / JYEAR_100  # jul cent. since J2000.0

        lambd = (218.32 + 481267.883 * t
                 + 6.29 * np.sin(np.deg2rad(134.9 + 477198.85 * t))
                 - 1.27 * np.sin(np.deg2rad(259.2 - 413335.38 * t))
                 + 0.66 * np.sin(np.deg2rad(235.7 + 890534.23 * t))
                 + 0.21 * np.sin(np.deg2rad(269.9 + 954397.70 * t))
                 - 0.19 * np.sin(np.deg2rad(357.5 + 35999.05 * t))
                 - 0.11 * np.sin(np.deg2rad(186.6 + 966404.05 * t)))
        self.lambd = np.deg2rad(lambd)

        beta = (5.13 * np.sin(np.deg2rad(93.3 + 483202.03 * t))
                + 0.28 * np.sin(np.deg2rad(228.2 + 960400.87 * t))
                - 0.28 * np.sin(np.deg2rad(318.3 + 6003.18 * t))
                - 0.17 * np.sin(np.deg2rad(217.6 - 407332.20 * t)))
        self.beta = np.deg2rad(beta)

        pie = (0.9508 + 0.0518 * np.cos(np.deg2rad(134.9 + 477198.85 * t))
               + 0.0095 * np.cos(np.deg2rad(259.2 - 413335.38 * t))
               + 0.0078 * np.cos(np.deg2rad(235.7 + 890534.23 * t))
               + 0.0028 * np.cos(np.deg2rad(269.9 + 954397.70 * t)))
        self.pie = np.deg2rad(pie)

    def _high_precision_calculations(self) -> None:
        """Compute accurate precession values for the moon location method calculations.
        """
        t = (self.time_ttjd - 2415020.) / JYEAR_100  # this based around 1900 ... */
        t_2 = t * t
        t_3 = t_2 * t
        lpr = 270.434164 + 481267.8831 * t - 0.001133 * t_2 + 0.0000019 * t_3
        m = 358.475833 + 35999.0498 * t - 0.000150 * t_2 - 0.0000033 * t_3
        mpr = 296.104608 + 477198.8491 * t + 0.009192 * t_2 + 0.0000144 * t_3
        d = 350.737486 + 445267.1142 * t - 0.001436 * t_2 + 0.0000019 * t_3
        f = 11.250889 + 483202.0251 * t - 0.003211 * t_2 - 0.0000003 * t_3
        om = 259.183275 - 1934.1420 * t + 0.002078 * t_2 + 0.0000022 * t_3

        lpr %= 360.
        mpr %= 360.
        m %= 360.
        d %= 360.
        f %= 360.
        om %= 360.

        sinx = np.sin(np.deg2rad(51.2 + 20.2 * t))
        lpr = lpr + 0.000233 * sinx
        m = m - 0.001778 * sinx
        mpr = mpr + 0.000817 * sinx
        d = d + 0.002011 * sinx

        sinx = 0.003964 * np.sin(np.deg2rad(346.560 + 132.870 * t - 0.0091731 * t_2))

        lpr = lpr + sinx
        mpr = mpr + sinx
        d = d + sinx
        f = f + sinx

        sinx = np.sin(np.deg2rad(om))
        lpr = lpr + 0.001964 * sinx
        mpr = mpr + 0.002541 * sinx
        d = d + 0.001964 * sinx
        f = f - 0.024691 * sinx
        f = f - 0.004328 * np.sin(np.deg2rad(om + 275.05 - 2.30 * t))

        e = 1 - 0.002495 * t - 0.00000752 * t_2

        m = np.deg2rad(m)  # these will all be arguments ... */
        mpr = np.deg2rad(mpr)
        d = np.deg2rad(d)
        f = np.deg2rad(f)

        lambd = (lpr + 6.288750 * np.sin(mpr)
                 + 1.274018 * np.sin(2 * d - mpr)
                 + 0.658309 * np.sin(2 * d)
                 + 0.213616 * np.sin(2 * mpr)
                 - e * 0.185596 * np.sin(m)
                 - 0.114336 * np.sin(2 * f)
                 + 0.058793 * np.sin(2 * d - 2 * mpr)
                 + e * 0.057212 * np.sin(2 * d - m - mpr)
                 + 0.053320 * np.sin(2 * d + mpr)
                 + e * 0.045874 * np.sin(2 * d - m)
                 + e * 0.041024 * np.sin(mpr - m)
                 - 0.034718 * np.sin(d)
                 - e * 0.030465 * np.sin(m + mpr)
                 + 0.015326 * np.sin(2 * d - 2 * f)
                 - 0.012528 * np.sin(2 * f + mpr)
                 - 0.010980 * np.sin(2 * f - mpr)
                 + 0.010674 * np.sin(4 * d - mpr)
                 + 0.010034 * np.sin(3 * mpr)
                 + 0.008548 * np.sin(4 * d - 2 * mpr)
                 - e * 0.007910 * np.sin(m - mpr + 2 * d)
                 - e * 0.006783 * np.sin(2 * d + m)
                 + 0.005162 * np.sin(mpr - d))

        lambd = (lambd + e * 0.005000 * np.sin(m + d)
                 + e * 0.004049 * np.sin(mpr - m + 2 * d)
                 + 0.003996 * np.sin(2 * mpr + 2 * d)
                 + 0.003862 * np.sin(4 * d)
                 + 0.003665 * np.sin(2 * d - 3 * mpr)
                 + e * 0.002695 * np.sin(2 * mpr - m)
                 + 0.002602 * np.sin(mpr - 2 * f - 2 * d)
                 + e * 0.002396 * np.sin(2 * d - m - 2 * mpr)
                 - 0.002349 * np.sin(mpr + d)
                 + e * e * 0.002249 * np.sin(2 * d - 2 * m)
                 - e * 0.002125 * np.sin(2 * mpr + m)
                 - e * e * 0.002079 * np.sin(2 * m)
                 + e * e * 0.002059 * np.sin(2 * d - mpr - 2 * m)
                 - 0.001773 * np.sin(mpr + 2 * d - 2 * f)
                 - 0.001595 * np.sin(2 * f + 2 * d)
                 + e * 0.001220 * np.sin(4 * d - m - mpr)
                 - 0.001110 * np.sin(2 * mpr + 2 * f)
                 + 0.000892 * np.sin(mpr - 3 * d)
                 - e * 0.000811 * np.sin(m + mpr + 2 * d)
                 + e * 0.000761 * np.sin(4 * d - m - 2 * mpr)
                 + e * e * 0.000717 * np.sin(mpr - 2 * m)
                 + e * e * 0.000704 * np.sin(mpr - 2 * m - 2 * d)
                 + e * 0.000693 * np.sin(m - 2 * mpr + 2 * d)
                 + e * 0.000598 * np.sin(2 * d - m - 2 * f)
                 + 0.000550 * np.sin(mpr + 4 * d)
                 + 0.000538 * np.sin(4 * mpr)
                 + e * 0.000521 * np.sin(4 * d - m)
                 + 0.000486 * np.sin(2 * mpr - d))

        b = (5.128189 * np.sin(f)
             + 0.280606 * np.sin(mpr + f)
             + 0.277693 * np.sin(mpr - f)
             + 0.173238 * np.sin(2 * d - f)
             + 0.055413 * np.sin(2 * d + f - mpr)
             + 0.046272 * np.sin(2 * d - f - mpr)
             + 0.032573 * np.sin(2 * d + f)
             + 0.017198 * np.sin(2 * mpr + f)
             + 0.009267 * np.sin(2 * d + mpr - f)
             + 0.008823 * np.sin(2 * mpr - f)
             + e * 0.008247 * np.sin(2 * d - m - f)
             + 0.004323 * np.sin(2 * d - f - 2 * mpr)
             + 0.004200 * np.sin(2 * d + f + mpr)
             + e * 0.003372 * np.sin(f - m - 2 * d)
             + 0.002472 * np.sin(2 * d + f - m - mpr)
             + e * 0.002222 * np.sin(2 * d + f - m)
             + e * 0.002072 * np.sin(2 * d - f - m - mpr)
             + e * 0.001877 * np.sin(f - m + mpr)
             + 0.001828 * np.sin(4 * d - f - mpr)
             - e * 0.001803 * np.sin(f + m)
             - 0.001750 * np.sin(3 * f)
             + e * 0.001570 * np.sin(mpr - m - f)
             - 0.001487 * np.sin(f + d)
             - e * 0.001481 * np.sin(f + m + mpr)
             + e * 0.001417 * np.sin(f - m - mpr)
             + e * 0.001350 * np.sin(f - m)
             + 0.001330 * np.sin(f - d)
             + 0.001106 * np.sin(f + 3 * mpr)
             + 0.001020 * np.sin(4 * d - f)
             + 0.000833 * np.sin(f + 4 * d - mpr))

        b = (b + 0.000781 * np.sin(mpr - 3 * f)
             + 0.000670 * np.sin(f + 4 * d - 2 * mpr)
             + 0.000606 * np.sin(2 * d - 3 * f)
             + 0.000597 * np.sin(2 * d + 2 * mpr - f)
             + e * 0.000492 * np.sin(2 * d + mpr - m - f)
             + 0.000450 * np.sin(2 * mpr - f - 2 * d)
             + 0.000439 * np.sin(3 * mpr - f)
             + 0.000423 * np.sin(f + 2 * d + 2 * mpr)
             + 0.000422 * np.sin(2 * d - f - 3 * mpr)
             - e * 0.000367 * np.sin(m + f + 2 * d - mpr)
             - e * 0.000353 * np.sin(m + f + 2 * d)
             + 0.000331 * np.sin(f + 4 * d)
             + e * 0.000317 * np.sin(2 * d + f - m + mpr)
             + e * e * 0.000306 * np.sin(2 * d - 2 * m - f)
             - 0.000283 * np.sin(mpr + 3 * f))

        om1 = 0.0004664 * np.cos(np.deg2rad(om))
        om2 = 0.0000754 * np.cos(np.deg2rad(om + 275.05 - 2.30 * t))

        beta = b * (1. - om1 - om2)

        self.pie = (0.950724 + 0.051818 * np.cos(mpr)
                    + 0.009531 * np.cos(2 * d - mpr)
                    + 0.007843 * np.cos(2 * d)
                    + 0.002824 * np.cos(2 * mpr)
                    + 0.000857 * np.cos(2 * d + mpr)
                    + e * 0.000533 * np.cos(2 * d - m)
                    + e * 0.000401 * np.cos(2 * d - m - mpr)
                    + e * 0.000320 * np.cos(mpr - m)
                    - 0.000271 * np.cos(d)
                    - e * 0.000264 * np.cos(m + mpr)
                    - 0.000198 * np.cos(2 * f - mpr)
                    + 0.000173 * np.cos(3 * mpr)
                    + 0.000167 * np.cos(4 * d - mpr)
                    - e * 0.000111 * np.cos(m)
                    + 0.000103 * np.cos(4 * d - 2 * mpr)
                    - 0.000084 * np.cos(2 * mpr - 2 * d)
                    - e * 0.000083 * np.cos(2 * d + m)
                    + 0.000079 * np.cos(2 * d + 2 * mpr)
                    + 0.000072 * np.cos(4 * d)
                    + e * 0.000064 * np.cos(2 * d - m + mpr)
                    - e * 0.000063 * np.cos(2 * d + m - mpr)
                    + e * 0.000041 * np.cos(m + d)
                    + e * 0.000035 * np.cos(2 * mpr - m)
                    - 0.000033 * np.cos(3 * mpr - 2 * d)
                    - 0.000030 * np.cos(mpr + d)
                    - 0.000029 * np.cos(2 * f - 2 * d)
                    - e * 0.000029 * np.cos(2 * mpr + m)
                    + e * e * 0.000026 * np.cos(2 * d - 2 * m)
                    - 0.000023 * np.cos(2 * f - 2 * d + mpr)
                    + e * 0.000019 * np.cos(4 * d - m - mpr))

        self.beta = Angle(np.deg2rad(beta), unit=u.rad)
        self.lambd = Angle(np.deg2rad(lambd), unit=u.rad)

    def low_precision_location(self, obs: EarthLocation) -> Tuple[SkyCoord, float]:
        """This is the same as the high precision method, but with a different set of coefficients.
        The difference is small. Good to about 0.1 deg, from the 1992 Astronomical Almanac, p. D46.
        Note that input time is a float.

        Args:
            obs: Earth location of the observation.

        Returns:
            Tuple[SkyCoord, float]: Moon coordinates and topographic distance.
        """
        self._low_precision_calculations()

        # Terrestrial time with julian day
        sid = local_sidereal_time(self.time, obs)
        lat = obs.lat
        distance = 1. / np.sin(self.pie)

        ell = np.cos(self.beta) * np.cos(self.lambd)
        m = 0.9175 * np.cos(self.beta) * np.sin(self.lambd) - 0.3978 * np.sin(self.beta)
        n = 0.3978 * np.cos(self.beta) * np.sin(self.lambd) + 0.9175 * np.sin(self.beta)

        x = ell * distance
        y = m * distance
        z = n * distance

        x = x - np.cos(lat) * np.cos(sid)
        y = y - np.cos(lat) * np.sin(sid)
        z = z - np.sin(lat)

        topo_dist = np.sqrt(x * x + y * y + z * z)

        ell = x / topo_dist
        m = y / topo_dist
        n = z / topo_dist

        alpha = np.arctan2(m, ell)
        delta = np.arcsin(n)
        distancemultiplier = Distance(EQUAT_RAD, unit=u.m)

        fr = current_geocent_frame(self.time)
        return SkyCoord(alpha, delta, topo_dist * distancemultiplier, frame=fr), topo_dist

    def accurate_location(self, obs: EarthLocation) -> Tuple[SkyCoord, float]:
        """Compute topocentric location and distance of moon to better accuracy.

        This is good to about 0.01 degrees.

        Args:
            obs: location on earth.

        Returns:
            Tuple[SkyCoord, float]: Moon coordinates and topographic distance.

        """
        self._high_precision_calculations()
        dist = Distance(1. / np.sin(np.deg2rad(self.pie)) * EQUAT_RAD)

        # place these in a skycoord in ecliptic coords of date.  Handle distance
        # separately since it does not transform properly for some reason.

        # eq = 'J{:7.2f}'.format(2000. + (time_ttjd[0] - _Constants.J2000) / JYEAR)
        equinox = f'J{2000. + (self.time_ttjd[0] - J2000) / JYEAR:7.2f}'
        frame = GeocentricTrueEcliptic(equinox=equinox)
        inecl = SkyCoord(lon=Angle(self.lambd, unit=u.rad), lat=Angle(self.beta, unit=u.rad), frame=frame)

        # Transform into geocentric equatorial.
        geocen = inecl.transform_to(current_geocent_frame(self.time))

        # Do the topo correction yourself. First form xyz coords in equatorial syst of date
        x = dist * np.cos(geocen.ra) * np.cos(geocen.dec)
        y = dist * np.sin(geocen.ra) * np.cos(geocen.dec)
        z = dist * np.sin(geocen.dec)

        # Now compute geocentric location of the observatory in a frame aligned with the
        # equatorial system of date, which one can do simply by replacing the west longitude
        # with the sidereal time

        # Exact match with thorskyutil/skycalc with the line below
        xobs, yobs, zobs = geocentric_coors(local_sidereal_time(self.time, obs), obs.lat, obs.height)

        # recenter moon's cartesian coordinates on position of obs
        x -= xobs
        y -= yobs
        z -= zobs

        # form the toposcentric ra and dec and bundle them into a skycoord of epoch of date.
        topo_dist = np.sqrt(x ** 2 + y ** 2 + z ** 2)
        raout = np.arctan2(y, x)
        decout = np.arcsin(z / topo_dist)

        if self.scalar_input:
            raout = np.squeeze(raout)
            decout = np.squeeze(decout)
            topo_dist = np.squeeze(topo_dist)

        return SkyCoord(raout, decout, unit=u.rad, frame=current_geocent_frame(self.time)), topo_dist

    @staticmethod
    def time_by_altitude(alt: Angle,
                         time_guess: Time,
                         location: EarthLocation) -> Optional[Time]:
        """Time at which moon passes a given altitude.

        This really does have to be iterated since the moon moves fairly
        quickly.

        Args:
            alt: If it is an array, then it must be the same length as time_guess desired altitude.
            time_guess : Initial guess; this needs to be fairly close.
            location : EarthLocation

        Returns:
            Optional[Time]: a Time, or None if non-convergent.
        """

        # time_guess is a Time, location is an EarthLocation

        time_guess = Time(np.asarray(time_guess.jd), format='jd')
        scalar_input = False
        if time_guess.ndim == 0:
            time_guess = time_guess[None]  # Makes 1D
            scalar_input = True
        alt = Angle(np.asarray(alt.to_value(u.rad)), unit=u.rad)
        if alt.ndim == 0:
            alt = alt[None]

        if len(time_guess) != len(alt):
            raise ValueError('Error: alt and guess must be the same length')

        moon = Moon()
        moon_pos, _ = moon.at(time_guess).accurate_location(location)
        tolerance = Angle(1.0e-4, unit=u.rad)

        delta = TimeDelta(0.002, format='jd')  # timestep

        ha = local_sidereal_time(time_guess, location) - moon_pos.ra

        alt2, az, parang = Altitude.above(moon_pos.dec, Angle(ha, unit=u.hourangle), location.lat)

        time_guess += delta
        moon_pos, _ = moon.at(time_guess).accurate_location(location)

        alt3, az, parang = Altitude.above(moon_pos.dec, local_sidereal_time(time_guess, location) - moon_pos.ra,
                                          location.lat)
        err = alt3 - alt

        deriv = (alt3 - alt2) / delta

        kount = np.zeros(len(time_guess), dtype=int)
        kk = np.where(np.logical_and(abs(err) > tolerance, kount < 10))[0][:]
        while len(kk) != 0:

            time_guess[kk] = time_guess[kk] - err[kk] / deriv[kk]

            moon_pos, _ = moon.at(time_guess[kk]).accurate_location(location)
            alt3[kk], az[kk], parang[kk] = Altitude.above(moon_pos.dec,
                                                          local_sidereal_time(time_guess[kk], location) - moon_pos.ra,
                                                          location.lat)
            err[kk] = alt3[kk] - alt[kk]
            ii = np.where(kount >= 9)[0][:]
            if len(ii) != 0:
                raise ValueError("Moonrise or set calculation not converging!\n")

            kk = np.where(np.logical_and(abs(err) > tolerance, kount < 10))[0][:]

        if scalar_input:
            time_guess = np.squeeze(time_guess)
        return Time(time_guess, format='iso')

    def rise_and_set(self,
                     location: EarthLocation,
                     midnight: Time,
                     set_alt: Angle,
                     rise_alt: Angle) -> Tuple[Time, Time]:
        """Calculates rise and set for the Moon

        Args:
            location (EarthLocation): Earth location of the observer
            midnight (Time): Midnight of that day as a Time object
            set_alt (Angle): Altitude for the moonset
            rise_alt (Angle): Altitude for the moonrise

        Returns:
            Tuple[Time, Time]: Moonrise and Moonset values, in that order.
        """
        moon_at_midnight, _ = self.at(midnight).low_precision_location(location)
        lst_midnight = local_sidereal_time(midnight, location)
        ha_moon_at_midnight = lst_midnight - moon_at_midnight.ra
        ha_moon_at_midnight.wrap_at(12. * u.hour, inplace=True)

        ha_moon_set = hour_angle_to_angle(moon_at_midnight.dec, location.lat, set_alt)  # corresponding hr angles
        diff_moon_set = ha_moon_set - ha_moon_at_midnight  # how far from setting point at midnight

        # find nearest setting point
        ii = np.where(diff_moon_set.hour >= 12.)[0][:]
        if len(ii) != 0:
            diff_moon_set[ii] = diff_moon_set[ii] - Angle(24. * u.hour)

        jj = np.where(diff_moon_set.hour < -12.)[0][:]
        if len(jj) != 0:
            diff_moon_set[jj] = diff_moon_set[jj] + Angle(24. * u.hour)

        timedelta_moon_set = TimeDelta(diff_moon_set.hour / 24., format='jd')
        times_moon_set = midnight + timedelta_moon_set
        times_moon_set = Moon.time_by_altitude(set_alt, times_moon_set, location)

        ha_moonrise = -1. * hour_angle_to_angle(moon_at_midnight.dec, location.lat, rise_alt)
        diff_moonrise = ha_moonrise - ha_moon_at_midnight  # how far from riseting point at midnight

        # find nearest riseting point
        ii = np.where(diff_moonrise.hour >= 12.)[0][:]
        if len(ii) != 0:
            diff_moonrise[ii] = diff_moonrise[ii] - Angle(24. * u.hour)
        jj = np.where(diff_moonrise.hour < -12.)[0][:]
        if len(jj) != 0:
            diff_moonrise[jj] = diff_moonrise[jj] + Angle(24. * u.hour)

        timedelta_moon_rise = TimeDelta(diff_moonrise.hour / 24., format='jd')
        times_moon_rise = midnight + timedelta_moon_rise
        times_moon_rise = Moon.time_by_altitude(rise_alt, times_moon_rise, location)

        return times_moon_rise, times_moon_set



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from typing import Optional, Tuple

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle, EarthLocation, SkyCoord
from astropy.time import Time, TimeDelta

from .altitude import Altitude
from .constants import J2000
from .utils import (current_geocent_frame, hour_angle_to_angle,
                    local_sidereal_time)


class Sun:
    """An interface to calculate different night events regarding the Sun.

    To use this is required to chain the `at` method at the beginning.
    If not unhandled errors would happen.
    """
    @staticmethod
    def at(time: Time) -> SkyCoord:
        """Low-precision position of the sun.

        Good to about 0.01 degree, from the 1990 Astronomical Almanac p. C24.
        At this level topocentric correction is not needed.

        Args:
            time: of the position

        Returns:
            SkyCoord: in the geocentric frame of epoch of date.

        """

        # Low precision formulae for the sun, from Almanac p. C24 (1990) */
        # said to be good to about a hundredth of a degree.

        jd = np.asarray(time.jd)
        scalar_input = False
        if jd.ndim == 0:
            jd = jd[None]
            scalar_input = True

        n = jd - J2000  # referred to J2000
        ell = 280.460 + 0.9856474 * n
        g = np.deg2rad(357.528 + 0.9856003 * n)
        lambd = np.deg2rad(ell + 1.915 * np.sin(g) + 0.020 * np.sin(2. * g))
        epsilon = np.deg2rad(23.439 - 0.0000004 * n)

        x = np.cos(lambd)
        y = np.cos(epsilon) * np.sin(lambd)
        z = np.sin(epsilon) * np.sin(lambd)

        ra = np.arctan2(y, x)
        dec = np.arcsin(z)

        frame = current_geocent_frame(time)

        if scalar_input:
            ra = np.squeeze(ra)
            dec = np.squeeze(dec)
        return SkyCoord(ra, dec, frame=frame, unit='radian')

    @staticmethod
    def time_by_altitude(alt: Angle,
                         time_guess: Time,
                         location: EarthLocation,
                         timestep: float = 0.002) -> Optional[Time]:
        """Time at which the sun crosses a particular altitude.

        This of course happens twice a day (or not at all);
        The usual use case will be to compute roughly when sunset or twilight occurs,
        and hand the result to this routine to get a more exact answer.

        time_guess is the starting time for iteration. This must be fairly close so that
        the iteration converges on the correct phenomenon (e.g., rise time, not set time).

        This uses the low-precision sun location, which is typically good to 0.01 degree.
        That's plenty good enough for computing rise, set, and twilight times.

        Args:
            alt: Desired altitude. If array, then must be the same length as time_guess.
            time_guess: Is a Time approximating the answer.
            location: EarthLocation
            timestep: The timestep to use.

        Raises:
            ValueError: Different lengths for Altitude and time_guess
            ArithmeticError: Sunrise, set, or twilight calculation not converging

        Returns:
            Time if convergent, None if non-convergent
        """
        time_guess = Time(np.asarray(time_guess.jd), format='jd')
        alt = Angle(np.asarray(alt.to_value(u.rad)), unit=u.rad)
        scalar_input = False
        if time_guess.ndim == 0 and alt.ndim == 0:
            scalar_input = True
        if time_guess.ndim == 0:
            time_guess = time_guess[None]  # Makes 1D
        if alt.ndim == 0:
            alt = alt[None]

        if len(time_guess) == 1 and len(alt) > 1:
            time_guess = Time(time_guess.jd * np.ones(len(alt)), format='jd')
        elif len(time_guess) > 1 and len(alt) == 1:
            alt *= np.ones(len(time_guess))
        elif len(time_guess) != len(alt):
            raise ValueError('Error: alt and time_guess have incompatible lengths')

        sun_pos = Sun.at(time_guess)
        tolerance = Angle(1.0e-4, unit=u.rad)

        delta = TimeDelta(timestep, format='jd')

        ha = local_sidereal_time(time_guess, location) - sun_pos.ra
        alt2, az, parang = Altitude.above(sun_pos.dec, Angle(ha, unit=u.hourangle), location.lat)

        time_guess += delta
        sun_pos = Sun.at(time_guess)

        alt3, az, parang = Altitude.above(sun_pos.dec,
                                          local_sidereal_time(time_guess, location) - sun_pos.ra,
                                          location.lat)
        err = alt3 - alt
        deriv = (alt3 - alt2) / delta

        kount = np.zeros(len(time_guess), dtype=int)
        kk = np.where(np.logical_and(abs(err) > tolerance, kount < 10))[0][:]
        while len(kk) != 0:
            time_guess[kk] = time_guess[kk] - err[kk] / deriv[kk]
            sun_pos = Sun.at(time_guess[kk])
            alt3[kk], az[kk], parang[kk] = Altitude.above(sun_pos.dec,
                                                          local_sidereal_time(time_guess[kk], location) - sun_pos.ra,
                                                          location.lat)
            err[kk] = alt3[kk] - alt[kk]
            kount[kk] += 1
            ii = np.where(kount >= 9)[0][:]
            if len(ii) != 0:
                raise ArithmeticError("Sunrise, set, or twilight calculation not converging.")
            kk = np.where(np.logical_and(abs(err) > tolerance, kount < 10))[0][:]

        if scalar_input:
            time_guess = np.squeeze(time_guess)
        return Time(time_guess, format='iso')

    @staticmethod
    def rise_and_set(location: EarthLocation,
                     time: Time,
                     midnight: Time,
                     set_alt: Angle,
                     rise_alt: Angle) -> Tuple[Time, Time, Time, Time]:
        """Compute rise and set times for this Sun.

            For the current location and time of the night.

        Args:
            location (EarthLocation): Earth location
            time (Time): time of the night.
            midnight (Time): Midnight time.
            set_alt (Angle): sunset altitude.
            rise_alt (Angle): sunrise altitude.

        Returns:
            The time of the event for the body in the precision of this Sun.
        """

        sun_at_midnight = Sun.at(midnight)
        lst_midnight = local_sidereal_time(midnight, location)
        nt = len(time)

        twelve_twilight_alt = Angle(-12. * np.ones(nt), unit=u.deg)  # 12 degree nautical twilight

        # corresponding hr angles
        sunset_ha = hour_angle_to_angle(sun_at_midnight.dec, location.lat, set_alt)
        sunrise_ha = Angle(2. * np.pi, unit=u.rad) - hour_angle_to_angle(sun_at_midnight.dec, location.lat, rise_alt)
        twelve_twilight_ha = hour_angle_to_angle(sun_at_midnight.dec, location.lat, twelve_twilight_alt)
        sun_at_midnight_ha = (lst_midnight - sun_at_midnight.ra).wrap_at(24. * u.hour)

        sunset_guess = sun_at_midnight_ha - sunset_ha  # angles away from midnight
        sunrise_guess = sunrise_ha - sun_at_midnight_ha
        even_12twi_guess = sun_at_midnight_ha - twelve_twilight_ha
        morn_12twi_guess = Angle(2. * np.pi, unit=u.rad) - twelve_twilight_ha - sun_at_midnight_ha

        # convert to time deltas
        timedelta_sunset = TimeDelta(sunset_guess.hour / 24., format='jd')
        timedelta_sunrise = TimeDelta(sunrise_guess.hour / 24., format='jd')
        timedelta_even_12twi = TimeDelta(even_12twi_guess.hour / 24., format='jd')
        timedelta_morn_12twi = TimeDelta(morn_12twi_guess.hour / 24., format='jd')

        # form into times and iterate to accurate answer.
        times_sunset = midnight - timedelta_sunset  # first approx
        times_sunset = Sun.time_by_altitude(set_alt, times_sunset, location)

        times_sunrise = midnight + timedelta_sunrise  # first approx
        times_sunrise = Sun.time_by_altitude(rise_alt, times_sunrise, location)

        times_even_12twi = midnight - timedelta_even_12twi
        times_even_12twi = Sun.time_by_altitude(twelve_twilight_alt, times_even_12twi, location)

        times_morn_12twi = midnight + timedelta_morn_12twi
        times_morn_12twi = Sun.time_by_altitude(twelve_twilight_alt, times_morn_12twi, location)

        return times_sunrise, times_sunset, times_even_12twi, times_morn_12twi



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum, IntEnum, auto
from typing import ClassVar, FrozenSet, List, Optional

from ..decorators import immutable
from ..types import ZeroTime
from .group import ROOT_GROUP_ID, AndGroup, Group
from .ids import ObservationID, ProgramID, UniqueGroupID
from .observation import Observation
from .semester import Semester
from .timeallocation import TimeAllocation
from .too import TooType


class Band(IntEnum):
    """
    Program band.
    """
    BAND1 = 1
    BAND2 = 2
    BAND3 = 3
    BAND4 = 4


class ProgramMode(IntEnum):
    """
    Main operational mode, which is one of:
    * Queue
    * Classical
    * Priority Visitor (hybrid mode between queue and classical)
    """
    QUEUE = auto()
    CLASSICAL = auto()
    PV = auto()


@immutable
@dataclass(frozen=True)
class ProgramType:
    """
    Represents the information encompassing the type of program.
    * abbreviation: the code used by the program type (e.g. Q, C, FT, LP)
    * name: user readable representation of the program type
    * is_science: indicates if this program type is a science program

    NOTE that ProgramType instances should NEVER be explicitly created.
    All of the valid ProgramType instances are contained in the ProgramTypes enum
    and should be accessed from there.
    """
    abbreviation: str
    name: str
    is_science: bool = True


class ProgramTypes(Enum):
    """
    A complete list of the ProgramType instances used by Gemini.
    As mentioned in ProgramType, ProgramType should never be instantiated
    outside of this enum: instead, ProgramType instances should be retrieved
    from here.
    """
    C = ProgramType('C', 'Classical')
    CAL = ProgramType('CAL', 'Calibration', False)
    DD = ProgramType('DD', "Director's Time")
    DS = ProgramType('DS', 'Demo Science')
    ENG = ProgramType('ENG', 'Engineering', False)
    FT = ProgramType('FT', 'Fast Turnaround')
    LP = ProgramType('LP', 'Large Program')
    Q = ProgramType('Q', 'Queue')
    SV = ProgramType('SV', 'System Verification')


@dataclass
class Program:
    """
    Representation of a program.

    The FUZZY_BOUNDARY is a constant that allows for a fuzzy boundary for a program's
    start and end times.
    """
    id: ProgramID
    internal_id: str
    # Some programs do not have a typical name and thus cannot be associated with a semester.
    semester: Optional[Semester]
    band: Band
    thesis: bool
    mode: ProgramMode
    type: Optional[ProgramTypes]
    start: datetime
    end: datetime
    allocated_time: FrozenSet[TimeAllocation] = field(hash=False, compare=False)

    # Root group is immutable and should not be used in hashing or comparisons.
    root_group: AndGroup

    too_type: Optional[TooType] = None

    FUZZY_BOUNDARY: ClassVar[timedelta] = timedelta(days=14)

    def __post_init__(self):
        if self.root_group.id != ROOT_GROUP_ID:
            raise ValueError(f"Program {self.id.id} should have root group should named {ROOT_GROUP_ID.id}, received: "
                             f'"{self.root_group.id}".')

    def program_awarded(self) -> timedelta:
        return sum((t.program_awarded for t in self.allocated_time), ZeroTime)

    def program_awarded_used(self) -> timedelta:
        return sum((t.program_used for t in self.allocated_time), ZeroTime)

    def partner_awarded(self) -> timedelta:
        return sum((t.partner_awarded for t in self.allocated_time), ZeroTime)

    def partner_awarded_used(self) -> timedelta:
        return sum((t.partner_used for t in self.allocated_time), ZeroTime)

    def total_awarded(self) -> timedelta:
        return sum((t.total_awarded() for t in self.allocated_time), ZeroTime)

    def total_awarded_used(self) -> timedelta:
        return sum((t.total_used() for t in self.allocated_time), ZeroTime)

    def program_used(self) -> timedelta:
        return self.root_group.program_used()

    def partner_used(self) -> timedelta:
        return self.root_group.partner_used()

    def total_used(self) -> timedelta:
        return self.root_group.total_used()

    def observations(self) -> List[Observation]:
        return self.root_group.observations()

    def get_group(self, group_id: UniqueGroupID) -> Optional[Group]:
        """
        Given a UniqueGroupID, find the subgroup with the ID if it exists.
        Returns None if no such group can be found.
        """
        def aux(group: Group) -> Optional[Group]:
            if group.unique_id == group_id:
                return group
            elif not isinstance(group.children, Observation):
                for subgroup in group.children:
                    retval = aux(subgroup)
                    if retval is not None:
                        return retval
            return None

        return aux(self.root_group)

    def get_observation(self, observation_id: ObservationID) -> Optional[Observation]:
        """
        Given an ObservationID, find the Observation with the ID if it exists.
        Returns None if no such Observation can be found.
        """
        def aux(group: Group) -> Optional[Observation]:
            match group.children:
                case Observation():
                    if group.children.id == observation_id:
                        return group.children
                    else:
                        return None
                case _:
                    for subgroup in group.children:
                        check_subgroup = aux(subgroup)
                        if check_subgroup is not None:
                            return check_subgroup
                    return None
        return aux(self.root_group)

    def show(self):
        """Print content of the Program.
        """
        print(f'Program: {self.id.id}')
        # Print the group and atom information.
        self.root_group.show(1)



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

"""
A small version of GPP data model
"""
from typing import NewType

from .atom import *
from .constraints import *
from .group import *
from .ids import *
from .magnitude import *
from .observation import *
from .observationmode import *
from .program import *
from .qastate import *
from .resource import *
from .semester import *
from .site import *
from .target import *
from .timeallocation import *
from .timingwindow import *
from .too import *
from .wavelength import *

# Type alias for a night index and night indices.
NightIndex = NewType('NightIndex', int)
NightIndices = npt.NDArray[NightIndex]

# Type alias for a time slot index and time slot indices.
TimeslotIndex = NewType('TimeslotIndex', int)
TimeslotIndices = npt.NDArray[TimeslotIndex]



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from typing import Tuple

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle, EarthLocation
from astropy.time import Time
from pytz import BaseTzInfo

from lucupy.types import TimeScalarOrNDArray

from .constants import EQUAT_RAD
from .moon import Moon
from .sun import Sun
from .utils import local_midnight_time


def night_events(time: Time, location: EarthLocation, localtzone: BaseTzInfo) -> Tuple[Time,
                                                                                       TimeScalarOrNDArray,
                                                                                       TimeScalarOrNDArray,
                                                                                       TimeScalarOrNDArray,
                                                                                       TimeScalarOrNDArray,
                                                                                       TimeScalarOrNDArray,
                                                                                       TimeScalarOrNDArray]:
    """Compute phenomena for a given night.
    This is mostly a testbed that prints results directly.

    Args:
        time : If before noon, events of previous night are computed.
        location: Earth Location
        localtzone: Timezone object.

    Returns:
        Tuple: A 7D tuple with either a ScalarOrNDArray (Check type for specific).

    """
    # prototype for the events of a single night -- sunset and rise,
    # twilights, and moonrise and set.

    time = Time(np.asarray(time.iso), format='iso')
    scalar_input = False
    if time.ndim == 0:
        time = time[None]  # Makes 1D
        scalar_input = True

    nt = len(time)

    midnight = local_midnight_time(time, localtzone)  # nearest clock-time midnight (UT)
    # allow separate rise and set altitudes for horizon effects
    horiz = (-0.883 - np.sqrt(2. * location.height / EQUAT_RAD) * (180. / np.pi)) * u.deg

    set_alt = Angle(horiz * np.ones(nt), unit=u.deg)  # zd = 90 deg 50 arcmin
    rise_alt = Angle(horiz * np.ones(nt), unit=u.deg)  # zd = 90 deg 50 arcmin

    # Sun
    sunrise, sunset, even_12twi, morn_12twi = Sun.rise_and_set(location, time, midnight, set_alt, rise_alt)

    # Moon
    moonrise, moonset = Moon().rise_and_set(location, midnight, set_alt, rise_alt)

    if scalar_input:
        sunset = np.squeeze(sunset)
        sunrise = np.squeeze(sunrise)
        even_12twi = np.squeeze(even_12twi)
        morn_12twi = np.squeeze(morn_12twi)
        moonrise = np.squeeze(moonrise)
        moonset = np.squeeze(moonset)

    return midnight, sunset, sunrise, even_12twi, morn_12twi, moonrise, moonset



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from __future__ import annotations

from abc import ABC
from datetime import timedelta
from typing import Optional, cast

from astropy.time import Time

from lucupy.minimodel.observationmode import ObservationMode, ObservationModes
from lucupy.minimodel.resource import Resource, Resources
from lucupy.minimodel.wavelength import Wavelengths


class ObservatoryProperties(ABC):
    """Observatory-specific methods.

       These are not tied to other components or
       structures, and allow computations to be implemented in one place.

    """
    _properties: Optional[ObservatoryProperties] = None

    @staticmethod
    def set_properties(cls) -> None:
        """Set properties for an specific Observatory

        Raises:
            ValueError: Illegal properties value.

        """
        if not issubclass(cls, ObservatoryProperties):
            raise ValueError('Illegal properties value.')
        ObservatoryProperties._properties = cls()

    @staticmethod
    def _check() -> ObservatoryProperties:
        """ Check if the properties have been set.

        Raises:
            ValueError: Properties have not been set.
        """
        if ObservatoryProperties._properties is None:
            raise ValueError('Observatory properties have not been set.')
        return cast(ObservatoryProperties, ObservatoryProperties._properties)

    @staticmethod
    def determine_standard_time(resources: Resources,
                                wavelengths: Wavelengths,
                                modes: ObservationModes,
                                cal_length: int) -> Time:
        """Determine standard time for a specific Observatory

        Args:
            resources (Resources): Set of Resources(instruments, mask, etc).
            wavelengths (Wavelengths): An array of Wavelengths to be observed.
            modes (ObservationModes): The different modes of observation.
            cal_length (int): The length (in seconds) of a calibration.

        Returns:
            Time: Value(s) of standard time
        """
        return ObservatoryProperties._check().determine_standard_time(
            resources,
            wavelengths,
            modes,
            cal_length
        )

    @staticmethod
    def is_instrument(resource: Resource) -> bool:
        """Determine if the given resource is an instrument or not.

        Args:
            resource (Resource): An instrument.

        Returns:
            bool: True is the resource is an instrument of the Observatory, otherwise False.
        """
        return ObservatoryProperties._check().is_instrument(resource)

    @staticmethod
    def acquisition_time(resource: Resource,
                         observation_mode: ObservationMode) -> Optional[timedelta]:
        """Given a resource, check if it is an instrument, and if so, lookup the
           acquisition time for the specified mode.

        Args:
            resource (Resource): A resource that should be an Instrument.
            observation_mode (ObservationMode): The observation mode to be used.

        Returns:
            Optional[timedelta]: The acquisition time for the instrument in that specific mode.
        """
        return ObservatoryProperties._check().acquisition_time(resource, observation_mode)



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from typing import FrozenSet, NewType

Wavelength = NewType('Wavelength', float)
Wavelengths = FrozenSet[float]



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

"""This will have to be customized by a given observatory if used independently
    of Gemini.
"""

from enum import Enum
from typing import Optional

import astropy.coordinates  # type: ignore
import pytz


class Site(Enum):
    """The sites belonging to the observatory using the Scheduler.

    Attributes:
        GN: Gemini North (568@399)
        GS: Gemini South (I11@399)

    """
    GN = ('Gemini North', '568@399')
    GS = ('Gemini South', 'I11@399')

    def __init__(self, site_name: str, coordinate_center: str, astropy_lookup: Optional[str] = None):
        self.site_name = site_name
        self.coordinate_center = coordinate_center

        if astropy_lookup is None:
            astropy_lookup = site_name.lower().replace(' ', '_')

        try:
            self.location = astropy.coordinates.EarthLocation.of_site(astropy_lookup)
        except astropy.coordinates.UnknownSiteException as e:
            msg = f'Unknown site lookup: {astropy_lookup}.'
            raise ValueError(e, msg)

        timezone_info = self.location.info.meta['timezone']
        try:
            self.timezone = pytz.timezone(timezone_info)
        except pytz.UnknownTimeZoneError as e:
            msg = f'Unknown time zone lookup: {timezone_info}.'
            raise ValueError(e, msg)


ALL_SITES = frozenset(s for s in Site)  # A variable to work with all the sites in scheduler components.



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

import numpy as np


def scalar_input(func):
    """
    Decorator to convert a function that returns a tuple to a function that returns a scalar.
    TODO: This is not currently being used. Ultimately, we would add it to the functions in the sky package.
    TODO: TO IMPLEMENT THIS IS NECESSARY TO REMOVE ASTROPY outputs or make them numpy compatible.
    """

    def wrapper(*args, **kwargs):
        # transform the input to numpy
        np_args = [np.asarray(arg) for arg in args]
        if any(arg.ndim == 0 for arg in np_args):
            args = [arg[np.newaxis] for arg in np_args]

        return np.squeeze(func(*args, **kwargs))

    return wrapper


def immutable(cls):
    """
    This marks a class as being immutable, i.e., when doing a copy or a deep copy, we only return self
    instead of actually making a copy of the class. It should only be used in dataclasses that are marked frozen.

    It works by implementing the __deepcopy__ and __copy__ dunder methods.
    """
    def __deepcopy__(self, _):
        return self

    def __copy__(self):
        return self

    setattr(cls, '__deepcopy__', __deepcopy__)
    setattr(cls, '__copy__', __copy__)
    return cls



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

"""Observatory is used to change behavior on some of the structures and components
   in the Scheduler service and others.

   Right now only Gemini Observatory is implemented but more should come.
   For new implementations, see Abstract.
"""



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from enum import IntEnum, auto


class QAState(IntEnum):
    """
    These correspond to the QA States in the OCS for Observations.
    Entries in the obs log should be made uppercase for lookups into
    this enum.

    PASS is not in original mini-model description, but returned by OCS.

    Members:
        - NONE
        - UNDEFINED
        - FAIL
        - USABLE
        - PASS
        - CHECK

    """
    NONE = auto()
    UNDEFINED = auto()
    FAIL = auto()
    USABLE = auto()
    PASS = auto()
    CHECK = auto()



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from abc import ABC
from dataclasses import dataclass
from enum import Enum, IntEnum, auto
from typing import NewType, NoReturn

import numpy.typing as npt

from ..decorators import immutable
from .magnitude import Magnitudes

TargetName = NewType('TargetName', str)


class TargetType(Enum):
    """The type associated with a target in an observation.

    Members:
        - BASE
        - USER
        - BLIND_OFFSET
        - OFF_AXIS
        - TUNING_STAR
        - GUIDESTAR
        - OTHER
    """
    BASE = auto()
    USER = auto()
    BLIND_OFFSET = auto()
    OFF_AXIS = auto()
    TUNING_STAR = auto()
    GUIDESTAR = auto()
    OTHER = auto()


class GuideSpeed(IntEnum):
    """
    How quickly a guider can guide on a guide star.

    Members:
        - SLOW
        - MEDIUM
        - FAST

    """
    SLOW = auto()
    MEDIUM = auto()
    FAST = auto()


class TargetTag(Enum):
    """
    A tag used by nonsidereal targets to indicate their type.
    """
    COMET = auto()
    ASTEROID = auto()
    MAJOR_BODY = auto()


@immutable
@dataclass(frozen=True)
class Target(ABC):
    """
    Basic target information.

    Attributes:
        - name: TargetName
        - magnitudes: Magnitudes
        - type: TargetType
    """
    name: TargetName
    magnitudes: Magnitudes
    type: TargetType

    def guide_speed(self) -> NoReturn:
        """
        Calculate the guide speed for this target.
        """
        raise NotImplementedError


@immutable
@dataclass(frozen=True)
class SiderealTarget(Target):
    """
    For a SiderealTarget, we have an RA and Dec and then proper motion information
    to calculate the exact position.

    RA and Dec should be specified in decimal degrees.
    Proper motion must be specified in milliarcseconds / year.
    Epoch must be the decimal year.

    NOTE: The proper motion adjusted coordinates can be found in the TargetInfo in coord.

    Attributes:
        ra (float): Right Ascension
        dec (float): Declination
        pm_ra (float): Proper motion of the right ascension component.
        pm_dec (float): Proper motion of the declination component.
        epoch (float): The epoch in which the ra / dec were measured.

    """
    ra: float
    dec: float
    pm_ra: float
    pm_dec: float
    epoch: float


@immutable
@dataclass(frozen=True)
class NonsiderealTarget(Target):
    """
    For a NonsiderealTarget, we have a HORIZONS designation to indicate the lookup
    information, a tag to determine the type of target, and arrays of ephemerides
    to specify the position.

    Attributes:
        des (str): Horizon designation
        tag (TargetTag): TargetTag
        ra (npt.NDArray[float]): Right Ascension
        dec (npt.NDArray[float]): Declination

    """
    des: str
    tag: TargetTag
    ra: npt.NDArray[float]
    dec: npt.NDArray[float]



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass
from enum import Enum, auto
from typing import FrozenSet, Optional

from ..decorators import immutable


class MagnitudeSystem(Enum):
    """
    List of magnitude systems associated with magnitude bands.

    Members:
       - VEGA
       - AB
       - JY
    """
    VEGA = auto()
    AB = auto()
    JY = auto()


@immutable
@dataclass(frozen=True)
class MagnitudeBand:
    """They are fully enumerated in MagnitudeBands, so they should be looked up by name there.
    THIS CLASS SHOULD NOT BE INSTANTIATED.

    Values for center and width are specified in microns.

    Attributes:
        name (str):
        center (float):
        width (float):
        system (MagnitudeSystem): Default to MagnitudeSystem.VEGA .
        description (str, optional): Default to None.

    """
    name: str
    center: float
    width: float
    system: MagnitudeSystem = MagnitudeSystem.VEGA
    description: Optional[str] = None


class MagnitudeBands(Enum):
    """
    It is unconventional to use lowercase characters in an enum, but to differentiate
    them from the uppercase magnitude bands, we must.

    Look up the MagnitudeBand from this Enum as follows:
    MagnitudeBands[name]

    """
    u = MagnitudeBand('u', 0.356, 0.046, MagnitudeSystem.AB, 'UV')
    g = MagnitudeBand('g', 0.483, 0.099, MagnitudeSystem.AB, 'green')
    r = MagnitudeBand('r', 0.626, 0.096, MagnitudeSystem.AB, 'red')
    i = MagnitudeBand('i', 0.767, 0.106, MagnitudeSystem.AB, 'far red')
    z = MagnitudeBand('z', 0.910, 0.125, MagnitudeSystem.AB, 'near-infrared')
    U = MagnitudeBand('U', 0.360, 0.075, description='ultraviolet')
    B = MagnitudeBand('B', 0.440, 0.090, description='blue')
    V = MagnitudeBand('V', 0.550, 0.085, description='visual')
    UC = MagnitudeBand('UC', 0.610, 0.063, description='UCAC')
    R = MagnitudeBand('R', 0.670, 0.100, description='red')
    I = MagnitudeBand('I', 0.870, 0.100, description='infrared')
    Y = MagnitudeBand('Y', 1.020, 0.120)
    J = MagnitudeBand('J', 1.250, 0.240)
    H = MagnitudeBand('H', 1.650, 0.300)
    K = MagnitudeBand('K', 2.200, 0.410)
    L = MagnitudeBand('L', 3.760, 0.700)
    M = MagnitudeBand('M', 4.770, 0.240)
    N = MagnitudeBand('N', 10.470, 5.230)
    Q = MagnitudeBand('Q', 20.130, 1.650)
    AP = MagnitudeBand('AP', 0.550, 0.085, description='apparent')


@immutable
@dataclass(frozen=True)
class Magnitude:
    """A magnitude value in a particular band.

    Attributes:
        band (MagnitudeBands):
        value (float):
        error (float): Default to None.
    """
    band: MagnitudeBands
    value: float
    error: Optional[float] = None


Magnitudes = FrozenSet[Magnitude]



from typing import List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from lucupy.minimodel.ids import ObservationID
from lucupy.minimodel.site import Site


def airmass(obs_airmass: npt.NDArray[float],
            obs_id: ObservationID,
            time_interval: Optional[npt.NDArray[int]] = None) -> None:
    """Plot airmass values for an observation given a specific time interval
         obs_airmass (npt.NDArray[float]): Array of airmass values.
         obs_id (ObservationID): Observation ID.
         time_interval: Optional[npt.NDArray[int]]: Interval of time to plot. Defaults to None if the
         plot is for the entire night.
    """

    x = np.array([i for i in range(len(obs_airmass))], dtype=int)
    p = plt.plot(obs_airmass)
    colour = p[-1].get_color()
    if time_interval is not None:
        plt.plot(x[time_interval], obs_airmass[time_interval], color=colour, linewidth=4)
    plt.ylim(2.5, 0.95)
    plt.title(obs_id.id)
    plt.xlabel('Time Slot')
    plt.ylabel('Airmass')
    plt.show()


def interval(score: npt.NDArray[int],
             interval_to_plot: npt.NDArray[int],
             best_interval: npt.NDArray[int],
             label: str = "") -> None:
    """Plot the interval on time slots vs score.
        score (npt.NDArray[int]): Array of scores to plot in the interval.
        interval_to_plot (npt.NDArray[int]): Array of indices indicating the time slots.
        best_interval (npt.NDArray[int]): Best time slots for that observation.
        label (str): Label for the plot. Defaults to an empty string.
    """

    x = np.array([i for i in range(len(score))], dtype=int)
    p = plt.plot(x, score)
    colour = p[-1].get_color()
    if best_interval is not None:
        plt.plot(x[best_interval], score[best_interval], color=colour, linewidth=4)

    plt.axvline(interval_to_plot[0], ymax=1.0, color='black')
    plt.axvline(interval_to_plot[-1], ymax=1.0, color='black')
    plt.ylabel('Score', fontsize=12)
    plt.xlabel('Time Slot', fontsize=12)
    if label != '':
        plt.title(label)
    plt.show()


def timelines(obs_order: List[Tuple[int, int, int]],
              sites: List[Site],
              obs_id: ObservationID,
              obs_airmass: npt.NDArray[int],
              scores: npt.NDArray[int],
              night: int = 0) -> None:
    """Airmass and Score vs time/slot plot of the timelines for a night.
        obs_order (List[Tuple[int, int, int]]): List of observations in order (with start and finish).
        sites (List[Site]): List of sites.
        obs_id (ObservationID):  Observation ID.
        obs_airmass (npt.NDArray[int]): Array of airmass values for the observation.
        scores (npt.NDArray[int]): Array of scores for the observation.
        night (int): Night index. Defaults to 0 or the first (or only) night.
    """

    for site in sites:
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=(10, 10))
        for idx, start, end in obs_order:
            if idx != -1:
                x = np.array([i for i in range(len(obs_airmass))], dtype=int)
                p = ax1.plot(x, obs_airmass)
                ax2.plot(x, np.log10(scores))

                colour = p[-1].get_color()
                ax1.plot(x[start:end + 1], obs_airmass[start:end + 1], linewidth=4, color=colour,
                         label=obs_id.id)
                ax2.plot(x[start:end + 1], np.log10(scores[start:end + 1]), linewidth=4, color=colour,
                         label=obs_id.id)

        ax1.axhline(2.0, xmax=1.0, color='black')
        ax1.set_ylim(2.5, 0.95)
        ax1.set_xlabel('Time Slot')
        ax1.set_ylabel('Airmass')
        ax1.set_title(f"Night {night + 1}: {site.name}")
        ax1.legend()

        ax2.set_xlabel('Time Slot')
        ax2.set_ylabel('log(Score)')
        ax2.set_title(f"Night {night + 1}: {site.name}")
        plt.show()



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import ClassVar, Optional

from ..decorators import immutable


@immutable
@dataclass(frozen=True)
class TimingWindow:
    """Representation of timing windows in the mini-model.

    Attributes:
        start (datetime): When a timing window begins.
        duration (timedelta): For infinite duration, set duration to timedelta.max.
        repeat (int):  -1 means forever repeating, 0 means non-repeating.
        period (timedelta, optional): None should be used if repeat < 1.
    """
    start: datetime
    duration: timedelta
    repeat: int
    period: Optional[timedelta]

    # For infinite duration, use the length of an LP.
    INFINITE_DURATION_FLAG: ClassVar[int] = -1
    INFINITE_DURATION: ClassVar[timedelta] = timedelta(days=3 * 365, hours=24)
    FOREVER_REPEATING: ClassVar[int] = -1
    NON_REPEATING: ClassVar[int] = 0
    NO_PERIOD: ClassVar[Optional[timedelta]] = None

    # A number to be used by the Scheduler to represent infinite repeats from the
    # perspective of the OCS: if FOREVER_REPEATING is selected, then it is converted
    # into this for calculation purposes.
    OCS_INFINITE_REPEATS: ClassVar[int] = 1000



# All code in this package is a refactored, numpy-vectorized version of thorskyutil.py:
#
# https://github.com/jrthorstensen/thorsky/blob/master/thorskyutil.py
#
# utility and miscellaneous time and the sky routines built mostly on astropy.
#
# Copyright John Thorstensen, 2018, who graciously has allowed Gemini to use this code under the BSD-3 Clause license.
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

"""
Constants that are used here and there.  Some are Quantities,
others are just floats. Not all are used.

The planet-coefs are for series expansions for the phase functions
of the planets, used in predicting apparent magnitude. See code.
"""

from typing import Final

import astropy.units as u
from astropy.coordinates.distances import Distance
from astropy.time import Time

PI: Final[float] = 3.14159265358979
TWOPI: Final[float] = 6.28318530717959
PI_OVER_2: Final[float] = 1.57079632679490
ARCSEC_IN_RADIAN: Final[float] = 206264.8062471
DEG_IN_RADIAN: Final[float] = 57.2957795130823
HRS_IN_RADIAN: Final[float] = 3.819718634205
KMS_AUDAY: Final[float] = 1731.45683633  # km per sec in 1 AU/day
SPEED_OF_LIGHT: Final[float] = 299792.458  # in km per sec ... exact.
SS_MASS: Final[float] = 1.00134198  # solar system mass in solar units
J2000: Final[float] = 2451545.  # Julian date at standard epoch
J2000_Time: Final[Time] = Time(2451545., format='jd')  # J2000 rendered as a Time
JYEAR: Final[float] = 365.25 # noqa
JYEAR_100: Final[float] = JYEAR * 100 # noqa
SEC_IN_DAY: Final[float] = 86400.
FLATTEN: Final[float] = 0.003352813  # flattening of earth, 1/298.257
EQUAT_RAD: Final[Distance] = 6378137. * u.m  # equatorial radius of earth, meters
EARTHRAD_IN_AU: Final[float] = 23454.7910556298  # number of earth rad in 1 au
ASTRO_UNIT: Final[float] = 1.4959787066e11  # 1 AU in meters
RSUN: Final[float] = 6.96000e8  # IAU 1976 recom. solar radius, meters
RMOON: Final[float] = 1.738e6  # IAU 1976 recom. lunar radius, meters
PLANET_TOL: Final[float] = 3.  # flag if nearer than 3 degrees
KZEN: Final[float] = 0.172  # mag / airmass relation for Hale Pohaku



# Copyright (c) 2016-2022 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause
""" # Lucupy

This package contains data structures and functions to support all the microservices that make up the Schedule app
for the Gemini Program Plataform (and other auxiliary services such as Env and Resource)

##Content

- Minimodel: A small version of GPP's data model. This allows for any services to use common data structures to model programs, observations, etc.
- Helpers: A collection of functions that helps with the handling of data.
- Observatory: An API that allows Observatory-specific behaviour to be added to a service (or sub-service: e.g Collector)
- Sky: A library that calculates night events for the use of visibility in Observations
- Time Utils: Time handling functions
- Types: A collection of variables to use with Python typing
- Decorators: A collection of decorators (Currently empty)

"""
__version__ = '0.1.3'



# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from enum import Enum, EnumMeta

import astropy.units as u
from astropy.time import Time

from lucupy.minimodel import (ObservationMode, ObservationModes, Resource,
                              Resources, Wavelengths)
from lucupy.observatory.abstract import ObservatoryProperties


class GeminiProperties(ObservatoryProperties):
    """Implementation of ObservatoryCalculations specific to Gemini.
    """

    class _InstrumentsMeta(EnumMeta):
        """Metaclass for the Instruments Class below.
        """
        def __contains__(cls, r: Resource) -> bool:  # type: ignore[override]
            return any(inst.value.id in r.id for inst in cls.__members__.values())  # type: ignore[var-annotated]

    class Instruments(Enum, metaclass=_InstrumentsMeta):
        """ Gemini-specific instruments.
        """
        FLAMINGOS2 = Resource('Flamingos2')
        NIFS = Resource('NIFS')
        NIRI = Resource('NIRI')
        IGRINS = Resource('IGRINS')
        GMOS_S = Resource('GMOS-S')
        GMOS_N = Resource('GMOS-N')
        GNIRS = Resource('GNIRS')
        GPI = Resource('GPI')
        GSAOI = Resource('GSAOI')

    _STANDARD_INSTRUMENTS = [Instruments.FLAMINGOS2,
                             Instruments.GNIRS,
                             Instruments.NIFS,
                             Instruments.IGRINS]
    """ List: Instruments for which there are set standards.
    """

    @staticmethod
    def determine_standard_time(resources: Resources,
                                wavelengths: Wavelengths,
                                modes: ObservationModes,
                                cal_length: int) -> Time:
        """Determine the standard star time required for Gemini.

        Args:
            resources (Resources): Instruments to be used.
            wavelengths (Wavelength): Wavelengths to be observed.
            modes (ObservationModes): Observation modes.
            cal_length (int): The specific length of a calibration.

        Returns:
            Time: _description_

        Todo:
            We may only want to include specific resources, in which case, modify
            Instruments above to be StandardInstruments.

        """
        if cal_length > 1:
            # Check to see if any of the resources are instruments.
            if any(resource in GeminiProperties._STANDARD_INSTRUMENTS for resource in resources):
                if all(wavelength <= 2.5 for wavelength in wavelengths):
                    return 1.5 * u.h
                else:
                    return 1.0 * u.h
            if ObservationMode.IMAGING in modes:
                return 2.0 * u.h
            return 0.0 * u.h

    @staticmethod
    def is_instrument(resource: Resource) -> bool:
        """Checks if the resource is a Gemini instrument.

        Args:
            resource (Resource): A resource to be checked.

        Returns:
            bool: True if resource is a Gemini instrument.
        """
        return resource in GeminiProperties.Instruments

    @staticmethod
    def acquisition_time(resource: Resource,
                         observation_mode: ObservationMode) -> None:
        """_summary_

        Args:
            resource (Resource): Instruments used.
            observation_mode (ObservationMode): Observation mode.

        Returns:
            Optional[timedelta]: _description_
        """
        if not GeminiProperties.is_instrument(resource):
            return None
        ...
